import logging
from logging import StreamHandler
import sys
import os

# #### LOGGER, MOVED FROM CONF BECAUSE OF MULTIPLE PROBLEM WITH CIRCULAR INCLUDES #####
CONAN_LOGGING_LEVEL = int(os.getenv('CONAN_LOGGING_LEVEL', logging.DEBUG))
CONAN_LOGGING_FILE = os.getenv('CONAN_LOGGING_FILE', None)  # None is stdout


class MultiLineFormatter(logging.Formatter):
    def format(self, record):
        str_ = logging.Formatter.format(self, record)
        separator = record.message if record.message else None
        if separator is None:
            return separator
        tmp = str_.split(separator)
        if len(tmp) == 2:
            header, _ = tmp
        else:
            header = tmp
#        num = len(header)
        num = 0
        str_ = str_.replace('\n', '\n' + ' ' * num)
        return str_


logger = logging.getLogger('conans')
if CONAN_LOGGING_FILE is not None:
    hdlr = logging.FileHandler(CONAN_LOGGING_FILE)
else:
    hdlr = StreamHandler(sys.stdout)

formatter = MultiLineFormatter('\n############## CONAN PACKAGE TOOLS ######################\n\n%(levelname)s: '
                               '%(message)s \n\n#########################################################\n')
hdlr.setFormatter(formatter)
for hand in logger.handlers:
    logger.removeHandler(hand)
logger.addHandler(hdlr)
logger.setLevel(CONAN_LOGGING_LEVEL)


#CRITICAL = 50
#FATAL = CRITICAL
#ERROR = 40
#WARNING = 30
#WARN = WARNING
#INFO = 20
#DEBUG = 10
#NOTSET = 0



import json
import os
import pipes
import platform
import tempfile
from collections import namedtuple

import sys

import shutil

from conan import __version__ as package_tools_version
from conan.log import logger
from conan.tools import get_bool_from_env
from conans.client.conan_api import Conan
from conans.client.profile_loader import _load_profile
from conans.model.version import Version
from conans.tools import vcvars_command
from conans.util.files import save, mkdir
from conans import __version__ as client_version


class TestPackageRunner(object):
    def __init__(self, profile_text, username, channel, reference,
                 mingw_installer_reference=None, runner=None,
                 args=None, conan_pip_package=None,
                 exclude_vcvars_precommand=False,
                 conan_vars=None, build_policy=None):

        self._conan_vars = conan_vars or {}
        self._profile_text = profile_text
        self._mingw_installer_reference = mingw_installer_reference
        self._args = args
        self._args += " --build=%s" % build_policy if build_policy else ""
        self._username = username
        self._channel = channel
        self._reference = reference
        self._conan_pip_package = conan_pip_package
        self._runner = runner or os.system
        self.runner = runner
        self.conan_api, self.client_cache, self.user_io = Conan.factory()
        self.conan_home = os.path.realpath(self.client_cache.conan_folder)
        self.data_home = os.path.realpath(self.client_cache.store)
        self._exclude_vcvars_precommand = exclude_vcvars_precommand

        if "default" in self._profile_text:  # User didn't specified a custom profile
            default_profile_name = os.path.basename(self.client_cache.default_profile_path)
            if not os.path.exists(self.client_cache.default_profile_path):
                self.conan_api.create_profile(default_profile_name, detect=True)

            if default_profile_name != "default":  # User have a different default profile name
                # https://github.com/conan-io/conan-package-tools/issues/121
                self._profile_text = self._profile_text.replace("include(default)",
                                                                "include(%s)" % default_profile_name)

        # Save the profile in a tmp file
        tmp = os.path.join(tempfile.mkdtemp(suffix='conan_package_tools_profiles'), "profile")
        self.abs_profile_path = os.path.abspath(tmp)
        save(self.abs_profile_path, self._profile_text)

        self.profile, _ = _load_profile(self._profile_text, os.path.dirname(self.abs_profile_path),
                                        self.client_cache.profiles_path)

    @property
    def settings(self):
        return self.profile.settings

    @property
    def options(self):
        return self.profile.options

    def run(self):
        pre_command = None
        compiler = self.settings.get("compiler", None)
        if not self._exclude_vcvars_precommand:
            if compiler == "Visual Studio" and "compiler.version" in self.settings:
                compiler_set = namedtuple("compiler", "version")(self.settings["compiler.version"])
                mock_sets = namedtuple("mock_settings",
                                       "arch compiler get_safe")(self.settings["arch"], compiler_set,
                                                                 lambda x: self.settings.get(x, None))
                pre_command = vcvars_command(mock_sets)

        self._run_create(pre_command=pre_command)

    def _run_create(self, pre_command=None):

        if Version(client_version) < Version("0.99"):  # Introduced in 1.0.0-beta.1
            path = ""
        else:
            path = "."

        ref = self._reference if self._reference else "%s/%s" % (self._username, self._channel)
        command = "conan create %s %s --profile %s %s" % (path, str(ref), self.abs_profile_path,
                                                          self._args)

        if pre_command:
            command = '%s && %s' % (pre_command, command)

        print("******** RUNNING BUILD ********** \n%s\n\n%s" % (command, self._profile_text))
        sys.stdout.flush()
        retcode = self._runner(command)
        if retcode != 0:
            exit("Error while executing:\n\t %s" % command)


def autodetect_docker_image(profile):
    compiler_name = profile.settings.get("compiler", None)
    compiler_version = profile.settings.get("compiler.version", None)
    if compiler_name not in ["clang", "gcc"]:
        raise Exception("Docker image cannot be autodetected for the compiler %s" % compiler_name)

    if compiler_name == "gcc" and Version(compiler_version) > Version("5"):
        compiler_version = Version(compiler_version).major(fill=False)

    return "lasote/conan%s%s" % (compiler_name, compiler_version.replace(".", ""))


class DockerTestPackageRunner(TestPackageRunner):
    def __init__(self, profile_text, username, channel, reference, mingw_ref=None, runner=None,
                 args=None, conan_pip_package=None, docker_image=None,
                 docker_image_skip_update=False, docker_32_images=False, build_policy=None):

        super(DockerTestPackageRunner, self).__init__(profile_text, username, channel, reference,
                                                      mingw_installer_reference=mingw_ref,
                                                      runner=runner, args=args,
                                                      conan_pip_package=conan_pip_package,
                                                      build_policy=build_policy)

        self.docker_image = docker_image or autodetect_docker_image(self.profile)
        if docker_32_images:
            self.docker_image += "-i386"
        self.docker_image_skip_update = docker_image_skip_update
        self.sudo_command = ""
        if "CONAN_DOCKER_USE_SUDO" in os.environ:
            if get_bool_from_env("CONAN_DOCKER_USE_SUDO"):
                self.sudo_command = "sudo"
        elif platform.system() == "Linux":
            self.sudo_command = "sudo"

    def clear_system_requirements(self):
        if self._reference:
            the_path = self.client_cache.system_reqs(self._reference)
            if os.path.exists(the_path):
                logger.info("Removing system_requirements: %s" % the_path)
                shutil.rmtree(os.path.dirname(the_path), ignore_errors=True)

    def run(self, pull_image=True, docker_entry_script=None):
        if pull_image:
            self.pull_image()
            if not self.docker_image_skip_update:
                # Update the downloaded image
                command = "%s docker run --name conan_runner %s /bin/sh -c " \
                        "\"sudo pip install conan_package_tools==%s " \
                        "--upgrade" % (self.sudo_command, self.docker_image, package_tools_version)
                if self._conan_pip_package:
                    command += " && sudo pip install %s\"" % self._conan_pip_package
                else:
                    command += " && sudo pip install conan --upgrade\""

                self._runner(command)
                # Save the image with the updated installed
                # packages and remove the intermediate container
                self._runner("%s docker commit conan_runner %s" % (self.sudo_command, self.docker_image))
                self._runner("%s docker rm conan_runner" % self.sudo_command)

        # Run the build
        serial = pipes.quote(self.serialize())
        env_vars = "-e CONAN_RUNNER_ENCODED=%s -e CONAN_USERNAME=%s " \
                   "-e CONAN_CHANNEL=%s" % (serial, self._username, self._channel)

        conan_env_vars = {key: value for key, value in os.environ.items()
                          if key.startswith("CONAN_") and key not in
                          ["CONAN_CHANNEL", "CONAN_USERNAME", "CONAN_USER_HOME"]}
        env_vars += " " + " ".join(['-e %s="%s"' % (key, value) for key, value in conan_env_vars.items()])

        command = "%s docker run --rm -v %s:/home/conan/project -v " \
                  "%s:/home/conan/.conan %s %s /bin/sh -c \"" \
                  "rm -f /home/conan/.conan/conan.conf && cd project && " \
                  "rm -f /home/conan/.conan/profiles/default && " \
                  "(conan profile new default --detect || true) && " \
                  "run_create_in_docker\"" % (self.sudo_command, os.getcwd(), self.conan_home,
                                              env_vars, self.docker_image)

        # Push entry command before to build
        if docker_entry_script:
            command = command.replace("run_create_in_docker", "%s && run_create_in_docker" % docker_entry_script)

        self.clear_system_requirements()
        ret = self._runner(command)
        if ret != 0:
            raise Exception("Error building: %s" % command)

    def pull_image(self):
        datadir = os.path.expanduser(self.data_home)
        if not os.path.exists(datadir):
            mkdir(datadir)
            if platform.system() != "Windows":
                self._runner("chmod -R 777 %s" % datadir)
        logger.info("Pulling docker image %s" % self.docker_image)
        self._runner("%s docker pull %s" % (self.sudo_command, self.docker_image))

    def serialize(self):
        doc = {"args": self._args,
               "username": self._username,
               "channel": self._channel,
               "profile": self._profile_text,
               "conan_pip_package": self._conan_pip_package,
               "reference": str(self._reference)}
        return json.dumps(doc)

    @staticmethod
    def deserialize(data):
        the_json = json.loads(data)
        ret = TestPackageRunner(the_json["profile"],
                                username=the_json["username"],
                                channel=the_json["channel"],
                                reference=the_json["reference"],
                                args=the_json["args"],
                                conan_pip_package=the_json["conan_pip_package"])
        return ret



import copy
from collections import namedtuple

from conans.model.ref import ConanFileReference


class BuildConf(namedtuple("BuildConf", "settings options env_vars build_requires reference")):

    def __new__(cls, settings, options, env_vars, build_requires, reference):
        if not isinstance(settings, dict):
            raise Exception("'settings' field has to be a dict")
        if not isinstance(options, dict):
            raise Exception("'options' field has to be a dict")
        if not isinstance(env_vars, dict):
            raise Exception("'env_vars' field has to be a dict")
        if not isinstance(build_requires, dict):
            raise Exception("'build_requires' field has to be a dict")
        if reference is not None and not isinstance(reference, str) and \
           not isinstance(reference, ConanFileReference):
            raise Exception("'reference' field has to be a string or ConanFileReference")

        if isinstance(reference, str):
            reference = ConanFileReference.loads(reference)

        return super(BuildConf, cls).__new__(cls, settings, options, env_vars,
                                             build_requires, reference)


def get_mingw_builds(mingw_configurations, mingw_installer_reference,
                     archs, shared_option_name, build_types, reference=None):
    builds = []
    for config in mingw_configurations:
        version, arch, exception, thread = config
        if arch not in archs:
            continue
        settings = {"arch": arch, "compiler": "gcc",
                    "compiler.version": version[0:3],
                    "compiler.threads": thread,
                    "compiler.exception": exception}
        build_requires = {"*": [mingw_installer_reference]}
        options = {}

        if shared_option_name:
            for shared in [True, False]:
                opt = copy.copy(options)
                opt[shared_option_name] = shared
                builds += _make_mingw_builds(settings, opt, build_requires, build_types, reference)
        else:
            builds += _make_mingw_builds(settings, options, build_requires, build_types, reference)

    return builds


def _make_mingw_builds(settings, options, build_requires, build_types, reference=None):
    builds = []

    for build_type_it in build_types:
        s2 = copy.copy(settings)
        s2.update({"build_type": build_type_it})
        s2.update({"compiler.libcxx": "libstdc++"})
        builds.append(BuildConf(s2, options, {}, build_requires, reference))

    return builds


def get_visual_builds(visual_versions, archs, visual_runtimes, shared_option_name,
                      dll_with_static_runtime, vs10_x86_64_enabled, build_types, reference=None):
    ret = []
    for visual_version in visual_versions:
        visual_version = str(visual_version)
        for arch in archs:
            if not vs10_x86_64_enabled and arch == "x86_64" and visual_version == "10":
                continue
            visual_builds = get_visual_builds_for_version(visual_runtimes, visual_version, arch,
                                                          shared_option_name,
                                                          dll_with_static_runtime, build_types,
                                                          reference)

            ret.extend(visual_builds)
    return ret


def get_visual_builds_for_version(visual_runtimes, visual_version, arch, shared_option_name,
                                  dll_with_static_runtime, build_types, reference=None):
    base_set = {"compiler": "Visual Studio",
                "compiler.version": visual_version,
                "arch": arch}
    sets = []

    if shared_option_name:
        if "MT" in visual_runtimes and "Release" in build_types:
            sets.append(({"build_type": "Release", "compiler.runtime": "MT"},
                         {shared_option_name: False}, {}, {}))
            if dll_with_static_runtime:
                sets.append(({"build_type": "Release", "compiler.runtime": "MT"},
                             {shared_option_name: True}, {}, {}))
        if "MTd" in visual_runtimes and "Debug" in build_types:
            sets.append(({"build_type": "Debug", "compiler.runtime": "MTd"},
                                  {shared_option_name: False}, {}, {}))
            if dll_with_static_runtime:
                sets.append(({"build_type": "Debug", "compiler.runtime": "MTd"},
                                      {shared_option_name: True}, {}, {}))
        if "MD" in visual_runtimes and "Release" in build_types:
            sets.append(({"build_type": "Release", "compiler.runtime": "MD"},
                                  {shared_option_name: False}, {}, {}))
            sets.append(({"build_type": "Release", "compiler.runtime": "MD"},
                                  {shared_option_name: True}, {}, {}))
        if "MDd" in visual_runtimes and "Debug" in build_types:
            sets.append(({"build_type": "Debug", "compiler.runtime": "MDd"},
                                  {shared_option_name: False}, {}, {}))
            sets.append(({"build_type": "Debug", "compiler.runtime": "MDd"},
                                  {shared_option_name: True}, {}, {}))

    else:
        if "MT" in visual_runtimes and "Release" in build_types:
            sets.append(({"build_type": "Release", "compiler.runtime": "MT"}, {}, {}, {}))
        if "MTd" in visual_runtimes and "Debug" in build_types:
            sets.append(({"build_type": "Debug", "compiler.runtime": "MTd"}, {}, {}, {}))
        if "MDd" in visual_runtimes and "Debug" in build_types:
            sets.append(({"build_type": "Debug", "compiler.runtime": "MDd"}, {}, {}, {}))
        if "MD" in visual_runtimes and "Release" in build_types:
            sets.append(({"build_type": "Release", "compiler.runtime": "MD"}, {}, {}, {}))

    ret = []
    for setting, options, env_vars, build_requires in sets:
        tmp = copy.copy(base_set)
        tmp.update(setting)
        ret.append(BuildConf(tmp, options, env_vars, build_requires, reference))

    return ret


def get_build(compiler, the_arch, the_build_type, the_compiler_version,
              the_libcxx, the_shared_option_name,
              the_shared, reference):
    options = {}
    if the_shared_option_name:
        options = {the_shared_option_name: the_shared}
    setts = {"arch": the_arch,
             "build_type": the_build_type,
             "compiler": compiler,
             "compiler.version": the_compiler_version}
    if the_libcxx:
        setts["compiler.libcxx"] = the_libcxx

    return BuildConf(setts, options, {}, {}, reference)


def get_osx_apple_clang_builds(apple_clang_versions, archs, shared_option_name,
                               pure_c, build_types, reference=None):
    ret = []
    # Not specified compiler or compiler version, will use the auto detected
    for compiler_version in apple_clang_versions:
        for arch in archs:
            if shared_option_name:
                for shared in [True, False]:
                    for build_type_it in build_types:
                        if not pure_c:
                            ret.append(get_build("apple-clang", arch, build_type_it,
                                                 compiler_version,
                                                 "libc++", shared_option_name, shared, reference))
                        else:
                            ret.append(get_build("apple-clang", arch, build_type_it,
                                                 compiler_version, None, shared_option_name,
                                                 shared, reference))
            else:
                for build_type_it in build_types:
                    if not pure_c:
                        ret.append(get_build("apple-clang", arch, build_type_it,
                                             compiler_version, "libc++", None, None, reference))
                    else:
                        ret.append(get_build("apple-clang", arch, build_type_it,
                                             compiler_version, None, None, None, reference))

    return ret


def get_linux_gcc_builds(gcc_versions, archs, shared_option_name, pure_c, build_types,
                         reference=None):
    ret = []
    # Not specified compiler or compiler version, will use the auto detected
    for gcc_version in gcc_versions:
        for arch in archs:
            if shared_option_name:
                for shared in [True, False]:
                    for build_type_it in build_types:
                        if not pure_c:
                            ret.append(get_build("gcc", arch, build_type_it, gcc_version,
                                                 "libstdc++", shared_option_name, shared,
                                                 reference))
                            if float(gcc_version) >= 5:
                                ret.append(get_build("gcc", arch, build_type_it, gcc_version,
                                                     "libstdc++11", shared_option_name, shared,
                                                     reference))
                        else:
                            ret.append(get_build("gcc", arch, build_type_it, gcc_version,
                                                 None, shared_option_name, shared, reference))
            else:
                for build_type_it in build_types:
                    if not pure_c:
                        ret.append(get_build("gcc", arch, build_type_it, gcc_version,
                                             "libstdc++", None, None, reference))
                        if float(gcc_version) >= 5:
                            ret.append(get_build("gcc", arch, build_type_it, gcc_version,
                                                 "libstdc++11", None, None, reference))
                    else:
                        ret.append(get_build("gcc", arch, build_type_it, gcc_version, None,
                                             None, None, reference))
    return ret


def get_linux_clang_builds(clang_versions, archs, shared_option_name, pure_c, build_types,
                           reference=None):
    ret = []
    # Not specified compiler or compiler version, will use the auto detected
    for clang_version in clang_versions:
        for arch in archs:
            if shared_option_name:
                for shared in [True, False]:
                    for build_type_it in build_types:
                        if not pure_c:
                            ret.append(get_build("clang", arch, build_type_it, clang_version,
                                                 "libstdc++", shared_option_name, shared,
                                                 reference))
                            ret.append(get_build("clang", arch, build_type_it, clang_version,
                                                 "libc++", shared_option_name, shared, reference))
                        else:
                            ret.append(get_build("clang", arch, build_type_it, clang_version,
                                                 None, shared_option_name, shared, reference))
            else:
                for build_type_it in build_types:
                    if not pure_c:
                        ret.append(get_build("clang", arch, build_type_it, clang_version,
                                             "libstdc++", None, None, reference))
                        ret.append(get_build("clang", arch, build_type_it, clang_version,
                                             "libc++", None, None, reference))
                    else:
                        ret.append(get_build("clang", arch, build_type_it, clang_version,
                                             None, None, None, reference))
    return ret
